/*
 * This code implements the MD5 message-digest algorithm.
 * The algorithm is due to Ron Rivest.	This code was
 * written by Colin Plumb in 1993, no copyright is claimed.
 * This code is in the public domain; do with it what you wish.
 *
 * Equivalent code is available from RSA Data Security, Inc.
 * This code has been tested against that, and is equivalent,
 * except that you don't need to include two pages of legalese
 * with every copy.
 *
 * To compute the message digest of a chunk of bytes, declare an
 * MD5Context structure, pass it to MD5Init, call MD5Update as
 * needed on buffers full of bytes, and then call MD5Final, which
 * will fill a supplied 16-byte array with the digest.
 */

/* Brutally hacked by John Walker back from ANSI C to K&R (no
   prototypes) to maintain the tradition that Netfone will compile
   with Sun's original "cc". */

/* Performance optimized by Paul Hsieh */

#include <memory.h>		 /* for memcpy() */
#include "md5.h"

/* Optionally include memory debugging mechanism */

#ifdef MEMORY_DEBUG
#include "memdbg.h"
#endif

#ifdef sgi
#define HIGHFIRST
#endif

#ifdef sun
#define HIGHFIRST
#endif

static void MD5Transform (uint32 buf[4], uint32 in[16]);

#ifndef HIGHFIRST
#define byteReverse(buf, len)	/* Nothing */
#else

#ifdef __WATCOMC__
static void byteReverse (unsigned char *buf, unsigned int longs) {
    uint32 t;
    do {
	t = *(uint32 *)buf;

	__asm {
		.486p
		mov	eax, t
		bswap	eax, eax
		mov	t, eax
	}

	*(uint32 *)buf = t;
	buf += 4;
    } while (--longs);
}
#else
/*
 * Note: this code is harmless on little-endian machines.
 */
static void byteReverse(buf, longs)
    unsigned char *buf; unsigned longs;
{
    uint32 t;
    do {
	t = (uint32) ((unsigned) buf[3] << 8 | buf[2]) << 16 |
	    ((unsigned) buf[1] << 8 | buf[0]);
	*(uint32 *) buf = t;
	buf += 4;
    } while (--longs);
}

#endif
#endif

/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
void MD5Init (struct MD5Context *ctx) {
    ctx->buf[0] = 0x67452301;
    ctx->buf[1] = 0xefcdab89;
    ctx->buf[2] = 0x98badcfe;
    ctx->buf[3] = 0x10325476;

    ctx->bits[0] = 0;
    ctx->bits[1] = 0;
}

/*
 * Update context to reflect the concatenation of another buffer full
 * of bytes.
 */
void MD5Update(struct MD5Context *ctx, const unsigned char *buf, unsigned len)
{
    uint32 t;

    /* Update bitcount */

    t = ctx->bits[0];
    if ((ctx->bits[0] = t + ((uint32) len << 3)) < t)
	ctx->bits[1]++; 	/* Carry from low to high */
    ctx->bits[1] += len >> 29;

    t = (t >> 3) & 0x3f;	/* Bytes already in shsInfo->data */

    /* Handle any leading odd-sized chunks */

    if (t) {
	unsigned char *p = (unsigned char *) ctx->in + t;

	t = 64 - t;
	if (len < t) {
	    memcpy(p, buf, len);
	    return;
	}
	/* The memcpy and byteReverse should be combined */
	memcpy(p, buf, t);
	byteReverse(ctx->in, 16);
	MD5Transform(ctx->buf, (uint32 *) ctx->in);
	buf += t;
	len -= t;
    }
    /* Process data in 64-byte chunks */

    while (len >= 64) {
	/* The memcpy and byteReverse should be combined */
	memcpy(ctx->in, buf, 64);
	byteReverse (ctx->in, 16);
	MD5Transform (ctx->buf, (uint32 *) ctx->in);
	buf += 64;
	len -= 64;
    }

    /* Handle any remaining bytes of data. */

    memcpy(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, MSB-first)
 */
void MD5Final(unsigned char digest[16], struct MD5Context *ctx)
{
    unsigned count;
    unsigned char *p;

    /* Compute number of bytes mod 64 */
    count = (ctx->bits[0] >> 3) & 0x3F;

    /* Set the first char of padding to 0x80.  This is safe since there is
       always at least one byte free */
    p = ctx->in + count;
    *p++ = 0x80;

    /* Bytes of padding needed to make 64 bytes */
    count = 64 - 1 - count;

    /* Pad out to 56 mod 64 */
    if (count < 8) {
	/* Two lots of padding:  Pad the first block to 64 bytes */
	memset(p, 0, count);
	byteReverse(ctx->in, 16);
	MD5Transform(ctx->buf, (uint32 *) ctx->in);

	/* Now fill the next block with 56 bytes */
	memset(ctx->in, 0, 56);
    } else {
	/* Pad block to 56 bytes */
	memset(p, 0, count - 8);
    }
    byteReverse(ctx->in, 14);

    /* Append length in bits and transform */
    ((uint32 *) ctx->in)[14] = ctx->bits[0];
    ((uint32 *) ctx->in)[15] = ctx->bits[1];

    MD5Transform(ctx->buf, (uint32 *) ctx->in);
    byteReverse((unsigned char *) ctx->buf, 4);
    memcpy(digest, ctx->buf, 16);
    memset(ctx, 0, sizeof(ctx));        /* In case it's sensitive */
}


/* The four core functions - F1 is optimized somewhat */

/* PH -- try to put off accessing to "x" as long as possible */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ ((y ^ z) & x))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (y ^ z ^ x)
#define F4(x, y, z) (y ^ (~z | x))

/* PH - Need to make sure the rol instruction is used */

#ifdef __WATCOMC__

uint32 ROL1 (uint32 w);
uint32 ROL2 (uint32 w);
uint32 ROL3 (uint32 w);
uint32 ROL4 (uint32 w);
uint32 ROL5 (uint32 w);
uint32 ROL6 (uint32 w);
uint32 ROL7 (uint32 w);
uint32 ROL8 (uint32 w);
uint32 ROL9 (uint32 w);
uint32 ROL10(uint32 w);
uint32 ROL11(uint32 w);
uint32 ROL12(uint32 w);
uint32 ROL13(uint32 w);
uint32 ROL14(uint32 w);
uint32 ROL15(uint32 w);
uint32 ROL16(uint32 w);
uint32 ROL17(uint32 w);
uint32 ROL18(uint32 w);
uint32 ROL19(uint32 w);
uint32 ROL20(uint32 w);
uint32 ROL21(uint32 w);
uint32 ROL22(uint32 w);
uint32 ROL23(uint32 w);
uint32 ROL24(uint32 w);
uint32 ROL25(uint32 w);
uint32 ROL26(uint32 w);
uint32 ROL27(uint32 w);
uint32 ROL28(uint32 w);
uint32 ROL29(uint32 w);
uint32 ROL30(uint32 w);
uint32 ROL31(uint32 w);

#pragma aux ROL1 = " rol eax, 1 " parm [eax] value [eax];
#pragma aux ROL2 = " rol eax, 2 " parm [eax] value [eax];
#pragma aux ROL3 = " rol eax, 3 " parm [eax] value [eax];
#pragma aux ROL4 = " rol eax, 4 " parm [eax] value [eax];
#pragma aux ROL5 = " rol eax, 5 " parm [eax] value [eax];
#pragma aux ROL6 = " rol eax, 6 " parm [eax] value [eax];
#pragma aux ROL7 = " rol eax, 7 " parm [eax] value [eax];
#pragma aux ROL8 = " rol eax, 8 " parm [eax] value [eax];
#pragma aux ROL9 = " rol eax, 9 " parm [eax] value [eax];
#pragma aux ROL10= " rol eax,10 " parm [eax] value [eax];
#pragma aux ROL11= " rol eax,11 " parm [eax] value [eax];
#pragma aux ROL12= " rol eax,12 " parm [eax] value [eax];
#pragma aux ROL13= " rol eax,13 " parm [eax] value [eax];
#pragma aux ROL14= " rol eax,14 " parm [eax] value [eax];
#pragma aux ROL15= " rol eax,15 " parm [eax] value [eax];
#pragma aux ROL16= " rol eax,16 " parm [eax] value [eax];
#pragma aux ROL17= " rol eax,17 " parm [eax] value [eax];
#pragma aux ROL18= " rol eax,18 " parm [eax] value [eax];
#pragma aux ROL19= " rol eax,19 " parm [eax] value [eax];
#pragma aux ROL20= " rol eax,20 " parm [eax] value [eax];
#pragma aux ROL21= " rol eax,21 " parm [eax] value [eax];
#pragma aux ROL22= " rol eax,22 " parm [eax] value [eax];
#pragma aux ROL23= " rol eax,23 " parm [eax] value [eax];
#pragma aux ROL24= " rol eax,24 " parm [eax] value [eax];
#pragma aux ROL25= " rol eax,25 " parm [eax] value [eax];
#pragma aux ROL26= " rol eax,26 " parm [eax] value [eax];
#pragma aux ROL27= " rol eax,27 " parm [eax] value [eax];
#pragma aux ROL28= " rol eax,28 " parm [eax] value [eax];
#pragma aux ROL29= " rol eax,29 " parm [eax] value [eax];
#pragma aux ROL30= " rol eax,30 " parm [eax] value [eax];
#pragma aux ROL31= " rol eax,31 " parm [eax] value [eax];

#define ROL0(w)  (w)
#define ROL32(w) (w)
#define ROL(w,s) ROL##s(w)

#else
#define ROL(w,s) ((w)<<(s) | (w)>>(32-(s)))
#endif

/* PH - schedule in the "+ data" earlier to remove a dependency on w */

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f, w, x, y, z, data, s) \
	( w += data,  w += f(x, y, z),  w = ROL (w, s),  w += x )

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
static void MD5Transform (uint32 buf[4], uint32 in[16]) {
    register uint32 a, b, c, d;

    a = buf[0];
    b = buf[1];
    c = buf[2];
    d = buf[3];

    MD5STEP (F1, a, b, c, d, in[ 0] + 0xd76aa478,  7);
    MD5STEP (F1, d, a, b, c, in[ 1] + 0xe8c7b756, 12);
    MD5STEP (F1, c, d, a, b, in[ 2] + 0x242070db, 17);
    MD5STEP (F1, b, c, d, a, in[ 3] + 0xc1bdceee, 22);
    MD5STEP (F1, a, b, c, d, in[ 4] + 0xf57c0faf,  7);
    MD5STEP (F1, d, a, b, c, in[ 5] + 0x4787c62a, 12);
    MD5STEP (F1, c, d, a, b, in[ 6] + 0xa8304613, 17);
    MD5STEP (F1, b, c, d, a, in[ 7] + 0xfd469501, 22);
    MD5STEP (F1, a, b, c, d, in[ 8] + 0x698098d8,  7);
    MD5STEP (F1, d, a, b, c, in[ 9] + 0x8b44f7af, 12);
    MD5STEP (F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
    MD5STEP (F1, b, c, d, a, in[11] + 0x895cd7be, 22);
    MD5STEP (F1, a, b, c, d, in[12] + 0x6b901122,  7);
    MD5STEP (F1, d, a, b, c, in[13] + 0xfd987193, 12);
    MD5STEP (F1, c, d, a, b, in[14] + 0xa679438e, 17);
    MD5STEP (F1, b, c, d, a, in[15] + 0x49b40821, 22);

    MD5STEP (F2, a, b, c, d, in[ 1] + 0xf61e2562,  5);
    MD5STEP (F2, d, a, b, c, in[ 6] + 0xc040b340,  9);
    MD5STEP (F2, c, d, a, b, in[11] + 0x265e5a51, 14);
    MD5STEP (F2, b, c, d, a, in[ 0] + 0xe9b6c7aa, 20);
    MD5STEP (F2, a, b, c, d, in[ 5] + 0xd62f105d,  5);
    MD5STEP (F2, d, a, b, c, in[10] + 0x02441453,  9);
    MD5STEP (F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
    MD5STEP (F2, b, c, d, a, in[ 4] + 0xe7d3fbc8, 20);
    MD5STEP (F2, a, b, c, d, in[ 9] + 0x21e1cde6,  5);
    MD5STEP (F2, d, a, b, c, in[14] + 0xc33707d6,  9);
    MD5STEP (F2, c, d, a, b, in[ 3] + 0xf4d50d87, 14);
    MD5STEP (F2, b, c, d, a, in[ 8] + 0x455a14ed, 20);
    MD5STEP (F2, a, b, c, d, in[13] + 0xa9e3e905,  5);
    MD5STEP (F2, d, a, b, c, in[ 2] + 0xfcefa3f8,  9);
    MD5STEP (F2, c, d, a, b, in[ 7] + 0x676f02d9, 14);
    MD5STEP (F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

    MD5STEP (F3, a, b, c, d, in[ 5] + 0xfffa3942,  4);
    MD5STEP (F3, d, a, b, c, in[ 8] + 0x8771f681, 11);
    MD5STEP (F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
    MD5STEP (F3, b, c, d, a, in[14] + 0xfde5380c, 23);
    MD5STEP (F3, a, b, c, d, in[ 1] + 0xa4beea44,  4);
    MD5STEP (F3, d, a, b, c, in[ 4] + 0x4bdecfa9, 11);
    MD5STEP (F3, c, d, a, b, in[ 7] + 0xf6bb4b60, 16);
    MD5STEP (F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
    MD5STEP (F3, a, b, c, d, in[13] + 0x289b7ec6,  4);
    MD5STEP (F3, d, a, b, c, in[ 0] + 0xeaa127fa, 11);
    MD5STEP (F3, c, d, a, b, in[ 3] + 0xd4ef3085, 16);
    MD5STEP (F3, b, c, d, a, in[ 6] + 0x04881d05, 23);
    MD5STEP (F3, a, b, c, d, in[ 9] + 0xd9d4d039,  4);
    MD5STEP (F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
    MD5STEP (F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
    MD5STEP (F3, b, c, d, a, in[ 2] + 0xc4ac5665, 23);

    MD5STEP (F4, a, b, c, d, in[ 0] + 0xf4292244,  6);
    MD5STEP (F4, d, a, b, c, in[ 7] + 0x432aff97, 10);
    MD5STEP (F4, c, d, a, b, in[14] + 0xab9423a7, 15);
    MD5STEP (F4, b, c, d, a, in[ 5] + 0xfc93a039, 21);
    MD5STEP (F4, a, b, c, d, in[12] + 0x655b59c3,  6);
    MD5STEP (F4, d, a, b, c, in[ 3] + 0x8f0ccc92, 10);
    MD5STEP (F4, c, d, a, b, in[10] + 0xffeff47d, 15);
    MD5STEP (F4, b, c, d, a, in[ 1] + 0x85845dd1, 21);
    MD5STEP (F4, a, b, c, d, in[ 8] + 0x6fa87e4f,  6);
    MD5STEP (F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
    MD5STEP (F4, c, d, a, b, in[ 6] + 0xa3014314, 15);
    MD5STEP (F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
    MD5STEP (F4, a, b, c, d, in[ 4] + 0xf7537e82,  6);
    MD5STEP (F4, d, a, b, c, in[11] + 0xbd3af235, 10);
    MD5STEP (F4, c, d, a, b, in[ 2] + 0x2ad7d2bb, 15);
    MD5STEP (F4, b, c, d, a, in[ 9] + 0xeb86d391, 21);
     
    buf[0] += a;
    buf[1] += b;
    buf[2] += c;
    buf[3] += d;
}
